class MessageableException(Exception):
    """
    Base class for user related errors.
    """
    message: str


class MissingPositional(MessageableException):
    """
    Raised when user forgets to pass positional.
    """

    def __init__(self, positional, position):
        """

        :param ezcli.Positional positional:
        :param int position:
        """
        self.positional = positional
        self.position: int = position
        self.message = f'Missing {position + 1}. argument!'


class TooManyValues(MessageableException):
    """
    Raised when user passes to many arguments. Require ``only_positionals = True`` in a ``Command`` class
    """

    def __init__(self, values, expected_len):
        """

        :param list[str] values:
        :param int expected_len:
        """
        self.message = f'Too many values, {len(values)} instead of {expected_len}'



import re


class Parameter:
    value: str = None
    """Value of parameter"""
    description: str = None
    """Description of param"""

    def pass_args(self, user_args: list[str]) -> list[str]:
        """
        This method gathers all args, and removes processed args.

        :param user_args: list of args to process
        :return: processed list of args, usally smaller then input
        """
        ...

    def __str__(self):
        return str(self.value)

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __bool__(self):
        return bool(self.value)


class Flag(Parameter):
    value: bool = False

    def __init__(self, *aliases: str):
        """

        :param aliases: must contains dash or dashes
        """
        self.aliases = aliases

    def pass_args(self, user_args: list[str]) -> list[str]:
        shrinked_args = [arg for arg in user_args if arg not in self.aliases]
        # noinspection PyTypeChecker
        self.value = len(shrinked_args) != len(user_args)
        return shrinked_args

    def __repr__(self):
        return f'<Flag {" ".join(self.aliases)} has value {self.value}>'


class KeyVal(Parameter):
    values: list[str] = []

    def __init__(self, keyname: str, default=None):
        """

        :param keyname: must contains dash or dashes
        """
        self.key = keyname
        if default is not None:
            self.values = [default]

    @property
    def value(self):
        try:
            return self.values[0]
        except IndexError:
            return None

    def pass_args(self, user_args: list[str]) -> list[str]:
        regex_rule = r'{}[=|\ ]\"?([^-][\w.,]*)\"?'.format(self.key)
        regex_match = re.findall(regex_rule, ' '.join(user_args))

        if not regex_match:
            return user_args

        self.values = list(regex_match)

        # Remove acquired values
        remove_tag = '<removeremoveremove>'
        h = user_args.copy()
        for i, arg in enumerate(h.copy()):
            if arg == self.key:
                h[i] = remove_tag
                continue
            h[i] = re.sub(regex_rule, remove_tag, arg)
        for i, arg in enumerate(h.copy()):
            if i > 0:
                if arg in self.values and h[i - 1] == remove_tag:
                    h[i] = remove_tag
        h = [k for k in h if k != remove_tag]

        return h

    def __repr__(self):
        return f'<Value of {self.key} has value {self.value}>'



from collections import OrderedDict
from typing import Optional

from colorama import Style, Back, Fore

from ezcliy.parameters import Flag, KeyVal


class HelpRenderer:
    def __init__(self, command):
        """

        :param ezcli.Command command:
        """
        self.command = command
        self.sections = OrderedDict()

    def add(self, section_name: str, val: str, desc: str):
        section = self.sections.get(section_name, [])
        section.append((val, desc))
        self.sections.update({section_name: section})

    @property
    def firstline(self):
        fline = Back.WHITE + Fore.BLACK + 'Usage:' + Style.RESET_ALL + f' {self.command.name}'

        if self.command.commands:
            fline += ' [COMMANDS]'

        if self.command.positionals:
            for pos in self.command.positionals:
                fline += f' [{pos.name}]'
        if not self.command.restrict_to_positionals_only:
            fline += f' [VALUES...]'

        if self.command.parameters:
            fline += ' [PARAMETERS]'
        return fline

    @staticmethod
    def render_section(section_name: str, rows: list[tuple[str, Optional[str]]]):
        lines = [Style.BRIGHT + f'{section_name.strip().capitalize()}:' + Style.RESET_ALL]
        for row in rows:
            line = f'    {row[0]}'
            if row[1] is not None:
                line += f' - {row[1]}'
            lines.append(line)
        return '\n'.join(lines)

    def __str__(self):
        lines = [self.firstline] + [self.render_section(n, r) for n, r in self.sections.items()]
        return '\n'.join(lines)


class Helpman(Flag):
    def __init__(self):
        super().__init__('--help')
        self.description = 'no description has been provided'

    def pass_args(self, user_args: list[str]) -> list[str]:
        if self.value:
            return user_args
        return super(Helpman, self).pass_args(user_args)

    @staticmethod
    def render_help(command):
        """

        :param ezcli.Command command:
        :return:
        """
        hr = HelpRenderer(command)

        # Get commands
        if command.commands:
            for n, c in command.commands.items():
                hr.add('commands', n, c().description)

        # Some positionals
        if command.positionals:
            for pos in command.positionals:
                hr.add('positionals', pos.name, pos.description)

        # Get params
        if command.parameters:
            for flag in [f for f in command.parameters.values() if isinstance(f, Flag)]:
                hr.add('flags', ' '.join(flag.aliases), flag.description)
            for keyval in [kv for kv in command.parameters.values() if isinstance(kv, KeyVal)]:
                hr.add('keyed values', keyval.key + ' (value)', keyval.description)

        return hr.__str__()



from typing import Optional

from ezcliy.exceptions import MissingPositional


class Positional:
    """
    Asign value (by source order) to object, allows asking for value or provide default one.
    """
    value: str = None
    """Fetched value by positional"""

    description: str = None
    """Description for help"""

    def __init__(self, name: str, ask_if_missing: Optional[str] = None, optional=False):
        self.name = name
        self.ask_if_missing = ask_if_missing
        self.optional = optional

    def pass_values(self, values: list[str], position: int):
        try:
            self.value = values[position]
        except IndexError:
            if self.ask_if_missing:
                self.value = input(self.ask_if_missing + ": ").strip()
            elif self.optional:
                pass
            else:
                raise MissingPositional(self, position)

    def __repr__(self):
        return f'<Positional expecting {self.name}>'

    def __str__(self):
        return self.value

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __bool__(self):
        return bool(self.value)



import asyncio
import inspect
import sys
from functools import cache

import colorama
import yaml

from ezcliy.exceptions import MessageableException, TooManyValues
from ezcliy.helpman import Helpman
from ezcliy.parameters import Parameter
from ezcliy.positional import Positional


class Command:
    """
    Base class for creating commands. You probably will override this class
    """
    name: str = None
    """Name of command."""

    description: str = None
    """Description for help"""

    values: list[str] = []
    """Cleaned values from cmd"""

    _legacy: dict[str, Parameter] = {}
    """Parameters inherited after previous command"""

    restrict_to_positionals_only = False
    """If true, allow only values referenced as positionals"""

    allow_empty_calls = False
    """If true, will not print help when command is issued without parameters"""

    _help: Helpman = None
    """Object that handles --help, it's just a powerful flag instance"""

    def __init__(self):
        if self.name is None:
            self.name = self.__class__.__name__.lower()
        if self.description is None:
            self.description = f'There is not description for {self.name}'
        self._help = Helpman()
        self.__subcommand_issued = None

    @property
    @cache
    def commands(self) -> dict[str, type]:
        """

        :return: Subcommands
        """
        cmds = dict()
        for pair in [{o().name: o} for o in self.__class__.__dict__.values() if isinstance(o, type) if
                     issubclass(o, Command)]:
            cmds.update(pair)
        return cmds

    @property
    def is_subcommand_issued(self):
        """

        :return: True if subcommand has been issued
        :rtype: bool
        """
        if self.__subcommand_issued is not None:
            return self.__subcommand_issued

        if len(self.values) > 0:
            self.__subcommand_issued = self.values[0] in self.commands.keys()
        else:
            self.__subcommand_issued = False
        return self.is_subcommand_issued

    @property
    @cache
    def parameters(self) -> dict[str, Parameter]:
        """

        :return: All declared parameters as name-parameter dict
        :rtype: dict[str, Parameter]
        """
        parameters = {'help': self._help}
        for name, param in [(n, p) for n, p in self.__class__.__dict__.items() if isinstance(p, Parameter)]:
            parameters.update({name: param})
        return parameters

    @property
    @cache
    def positionals(self):
        """

        :return: All declared positionals
        :rtype: list[Positional]
        """
        return [p for p in self.__class__.__dict__.values() if isinstance(p, Positional)]

    def __help_check(self):
        if self._help:
            print(self._help.render_help(self))
            exit()

        if len([p for p in self.positionals if p is not None]):
            return

    def __restriction_check(self):
        if len(self.values) != len(self.positionals) and self.restrict_to_positionals_only:
            raise TooManyValues(self.values, len(self.positionals))

    def dispatch(self, args: list[str]):
        """
        The magic spaghetti of framework.

        :param args: list of arguments to process
        """
        # Loading predecessor's parameters
        for n in [name for name, t in self.__annotations__.items() if isinstance(t, type) if issubclass(t, Parameter)]:
            if n in self._legacy:
                self.__setattr__(n, self._legacy.get(n))

        # Special help check
        if not len(args) and not self.allow_empty_calls:
            print(self._help.render_help(self))
            exit()


        # Shrink requested parameters
        for argument, name in [(self.parameters.get(a), a) for a in self.parameters]:
            args = argument.pass_args(args)

        # Save cleaned values
        self.values = [arg for arg in args if not arg.startswith('-') and arg != '']

        def pass_values_to_positionals():
            for i, p in enumerate(self.positionals):
                p.pass_values(self.values, i)

        # Check is first arg an command
        if len(self.values) > 0:
            if self.is_subcommand_issued:
                # Shrink values
                cmd_name, self.values = self.values[0], self.values[1:]
                # Checks
                self.__restriction_check()
                # Invoke, has args, first is subcommand
                if not self._help:
                    self.__invoke_handler()
                # Prepare subcommand
                cmd: Command = self.commands.get(cmd_name)()
                cmd._legacy = self.parameters.copy()  # Passing aquired flags
                cmd.dispatch(args[args.index(cmd_name) + 1:])  # Passing only args after command
                return

        self.__help_check()
        self.__restriction_check()
        pass_values_to_positionals()  # This should raise an error
        self.__invoke_handler()  # Has no args

    def __invoke_handler(self):
        if inspect.iscoroutinefunction(self.invoke):
            # noinspection PyTypeChecker
            asyncio.run(self.invoke())
        else:
            self.invoke()

    # Invocation
    def invoke(self):
        """
        The entry point for your command.
        """
        ...

    # Entry points
    def entry(self, *args: str):
        """
        Handles errors caused by dispatch method. Used also for testing. Use it on dev env.

        :param str args: list of passed arguments
        """
        try:
            self.dispatch(list(args))
        except MessageableException as mex:
            print(colorama.Fore.RED +
                  yaml.safe_dump({
                      'error': mex.__class__.__name__,
                      'message': mex.message
                  }) + colorama.Style.RESET_ALL
                  )

    def cli_entry(self):
        """
        This method grabs args from terminal. You should put it into ``if main``. This should be used in production.
        """
        self.entry(*sys.argv[1:])



from setuptools import setup
from ezcliy import __version__


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    install_requires=["colorama==0.4.4", "pyyaml==5.4.1"],
    name="ezcliy",
    version=__version__,
    packages=["ezcliy"],
    project_urls={
        'Repository': 'https://github.com/kpostekk/ezcliy/',
        'Docs': 'https://ezcliy.readthedocs.io/en/latest/'
    },
    url="https://github.com/kpostekk/ezcliy",
    license="Apache",
    author="Krystian Postek",
    author_email="krystian@postek.eu",
    description="Framework for creating CLI tools",
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    long_description=readme(),
    long_description_content_type="text/markdown",
    python_requires=">=3.9"
)
